package com.humaorie.dollar;

/*
 * Dollar, http://bitbucket.org/dfa/dollar
 * (c) 2010 Davide Angelocola <davide.angelocola@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * The single entry-point of the whole Dollar API. Usage:
 *
 * <code>
 * import static com.humaorie.dollar.Dollar.*;
 * </code>
 */
public final class Dollar {

    private Dollar() {
    }

    // wrapper
    public static interface Wrapper<T> extends Iterable<T> {

        // read-only: no mutations 
        int size();

        String join();

        String join(String separator);

        Wrapper<T> repeat(int n);

        Wrapper<T> shuffle();

        Wrapper<T> shuffle(Random random);

        Wrapper<T> copy();

        Wrapper<T> concat(Wrapper<T> wrapper);

        Wrapper<T> concat(T... items);

        Wrapper<T> concat(Collection<T> items);

        Wrapper<T> reverse();

        Wrapper<T> fill(T object);

        // if j > 0 returns the first j elements
        // if j < 0 returns the last j elements
        // if j == 0 does nothing
        // if |j| > len does nothing
        // alternatively slice(j) can remove the first j elements
        // and slice(-j) can remove the last j elements
        Wrapper<T> slice(int j);

        // returns all but the first i elements and the last j elements
        Wrapper<T> slice(int i, int j);

        Wrapper<T> sort();

        List<T> toList();

        @SuppressWarnings("unchecked")
		List<T> toList(Class<? extends List> concreteListClass);

        Set<T> toSet();

        @SuppressWarnings("unchecked")
		Set<T> toSet(Class<? extends Set> concreteSetClass);

        T[] toArray();

        // suggested by sbi
        // use when all elements are null, type guessing is useless
        // or use when you want to be explicit
        T[] toArray(Class<T> concreteArrayClass);

        ArrayWrapper<T> convert();

        Wrapper<T> filter(Function<Boolean, T> predicate);

        <R> Wrapper<R> map(Function<R, T> function);
    }

    public static interface Function<R, T> {
        
        R call(T object);
    }

    // Array wrapper
    public static interface ArrayWrapper<T> extends Wrapper<T> {

        byte[] toByteArray();

        short[] toShortArray();

        int[] toIntArray();

        long[] toLongArray();

        char[] toCharArray();

        boolean[] toBooleanArray();

        float[] toFloatArray();

        double[] toDoubleArray();
    }

    public static <T> ArrayWrapper<T> $(T[] array) {
        return new com.humaorie.dollar.ArrayWrapper<T>(array);
    }

    public static ArrayWrapper<Character> $(char[] components) {
        return new CharacterArrayWrapper(components);
    }

    public static ArrayWrapper<Boolean> $(boolean[] components) {
        return new BooleanArrayWrapper(components);
    }

    public static ArrayWrapper<Byte> $(byte[] components) {
        return new ByteArrayWrapper(components);
    }

    public static ArrayWrapper<Short> $(short[] components) {
        return new ShortArrayWrapper(components);
    }

    public static ArrayWrapper<Integer> $(int[] components) {
        return new IntegerArrayWrapper(components);
    }

    public static ArrayWrapper<Long> $(long[] components) {
        return new LongArrayWrapper(components);
    }

    public static ArrayWrapper<Float> $(float[] components) {
        return new FloatArrayWrapper(components);
    }

    public static ArrayWrapper<Double> $(double[] components) {
        return new DoubleArrayWrapper(components);
    }

    // List wrapper
    public static interface ListWrapper<T> extends Wrapper<T> {

        @Override
        ListWrapper<T> copy();

        ListWrapper<T> add(T item);

        ListWrapper<T> addAll(Iterable<T> items);

        ListWrapper<T> threadSafe();

        ListWrapper<T> immutable();

        ListWrapper<T> checked(Class<T> requiredClass);
    }

    public static <T> ListWrapper<T> $(List<T> list) {
        return new com.humaorie.dollar.ListWrapper<T>(list);
    }

    // Set wrapper
    public static interface SetWrapper<T> extends Wrapper<T> {

        @Override
        SetWrapper<T> copy();

        SetWrapper<T> add(T item);

        SetWrapper<T> addAll(Iterable<T> items);

        SetWrapper<T> threadSafe();

        SetWrapper<T> immutable();

        SetWrapper<T> checked(Class<T> requiredClass);
    }

    public static <T> SetWrapper<T> $(Set<T> set) {
        return new com.humaorie.dollar.SetWrapper<T>(set);
    }

    // Map wrapper
    public static interface MapWrapper<K, V> extends Wrapper<K> {

        MapWrapper<K, V> add(K key, V value);

        MapWrapper<K, V> threadSafe();

        MapWrapper<K, V> immutable();

        MapWrapper<K, V> checked(Class<K> requiredKeyClass, Class<V> requiredValueClass);

        Map<K, V> toMap();
    }

    public static <K, V> MapWrapper<K, V> $(Map<K, V> map) {
        return new com.humaorie.dollar.MapWrapper<K, V>(map);
    }

    // range iterators
    public static Wrapper<Integer> $(int stopAt) {
        return new RangeWrapper(stopAt);
    }

    public static Wrapper<Integer> $(int startFrom, int stopAt) {
        return new RangeWrapper(startFrom, stopAt);
    }

    public static Wrapper<Integer> $(int startFrom, int stopAt, int stepBy) {
        return new RangeWrapper(startFrom, stopAt, stepBy);
    }

    // range of characters, like A-Z
    public static Wrapper<Character> $(char from, char to) {
        return new CharRangeWrapper(from, to);
    }

    // String, StringBuilder, etc as an Iterable
    public static Wrapper<Character> $(CharSequence charSequence) {
        return new CharSequenceWrapper(charSequence);
    }

    // DateWrapper
    public static Wrapper<Date> $(Date begin, Date end) {
        return new DateRangeWrapper(begin, end);
    }

    // make an iterator iterable
    public static <T> Iterable<T> $(final Iterator<T> anIterator) {
        return new Iterable<T>() {

            @Override
            public Iterator<T> iterator() {
                return anIterator;
            }
        };
    }

    // special case: NullWrapper
    public static <T> Wrapper<T> $() {
        return new NullWrapper<T>();
    }

    // random data generators
    public static interface RandomWrapper<T> extends Wrapper<T> {

        RandomWrapper<T> samples(int samples);

        RandomWrapper<T> generator(Random random);
    }

    public static RandomWrapper<Float> $(Random random, float upTo) {
        return new RandomFloatWrapper(random, 1, upTo);
    }

    public static RandomWrapper<Double> $(Random random, double upTo) {
        return new RandomDoubleWrapper(random, 1, upTo);
    }

    public static RandomWrapper<Integer> $(Random random, int upTo) {
        return new RandomIntegerWrapper(random, 1, upTo);
    }

    public static RandomWrapper<Long> $(Random random, long upTo) {
        return new RandomLongWrapper(random, 1, upTo);
    }

    public static RandomWrapper<Boolean> $(Random random) {
        return new RandomBooleanWrapper(random, 1);
    }
}
