package com.humaorie.dollar;

/*
 * Dollar, http://bitbucket.org/dfa/dollar
 * (c) 2010 Davide Angelocola <davide.angelocola@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
import com.humaorie.dollar.Dollar.Wrapper;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Random;

class RangeWrapper extends AbstractWrapper<Integer> {

    private final int startFrom;
    private final int stopAt;
    private final int stepBy;

    public RangeWrapper(int stopAt) {
        if (stopAt < 0) {
            this.startFrom = stopAt;
            this.stopAt = 0;
            this.stepBy = 1;
        } else {
            this.startFrom = 0;
            this.stopAt = stopAt;
            this.stepBy = 1;
        }
    }

    public RangeWrapper(int startFrom, int stopAt) {
        if (startFrom > stopAt) {
            this.stepBy = -1;
            stopAt--;
            startFrom--;
        } else {
            this.stepBy = 1;
        }

        this.startFrom = startFrom;
        this.stopAt = stopAt;
    }

    public RangeWrapper(int startFrom, int stopAt, int stepBy) {
        if (startFrom > stopAt && stepBy > 0) { // desc loop
            this.startFrom = 1;
            this.stopAt = 1;
            this.stepBy = 1;
        } else if (stopAt > startFrom && stepBy < 0) { // asc loop
            this.startFrom = 1;
            this.stopAt = 1;
            this.stepBy = 1;
        } else if (stepBy == 0) { // cannot step by 0!
            this.startFrom = 1;
            this.stopAt = 1;
            this.stepBy = 1;
        } else {
            this.startFrom = startFrom;
            this.stopAt = stopAt;
            this.stepBy = stepBy;
        }
    }

    @Override
    public Wrapper<Integer> copy() {
        return new RangeWrapper(startFrom, stopAt, stepBy);
    }

    @Override
    public Wrapper<Integer> reverse() {
        // XXX: buggy!
        //        if (stepBy > 0) {
        //            return new RangeWrapper(stopAt - 1, startFrom - 1, -stepBy);
        //        } else {
        //            return new RangeWrapper(startFrom , stopAt , stepBy);
        //        }
        return new ListWrapper<Integer>(this).reverse();
    }

    @Override
    public Wrapper<Integer> shuffle(Random random) {
        return new ArrayWrapper<Integer>(copy().toArray()).shuffle(random);
    }

    @Override
    public int size() {
        return Math.abs(startFrom - stopAt) / Math.abs(stepBy);
    }

    @Override
    public Wrapper<Integer> sort() {
        if (stepBy > 0) {
            return this;
        } else {
            return new RangeWrapper(startFrom, stopAt, stepBy);
        }
    }

    @Override
    public Iterator<Integer> iterator() {
        if (stepBy > 0) {
            return new RangeIterator();
        } else {
            return new ReversedRangeIterator();
        }
    }

    class RangeIterator implements Iterator<Integer> {

        protected int i;

        public RangeIterator() {
            i = startFrom;
        }

        @Override
        public boolean hasNext() {
            return i < stopAt;
        }

        @Override
        public Integer next() {
            if (!hasNext()) {
                throw new NoSuchElementException();
            }

            int tmp = i;
            i += stepBy;
            return tmp;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("remove() not supported for RangeWrapper");
        }
    }

    class ReversedRangeIterator extends RangeIterator {

        @Override
        public boolean hasNext() {
            return i > stopAt;
        }
    }

    @Override
    public boolean equals(Object object) {
        if (object instanceof RangeWrapper) {
            RangeWrapper rangeWrapper = (RangeWrapper) object;
            return startFrom == rangeWrapper.startFrom
                    && stopAt == rangeWrapper.stopAt
                    && stepBy == rangeWrapper.stepBy;
        }

        return false;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash *= 19 + startFrom;
        hash *= 19 + stopAt;
        hash *= 19 + stepBy;
        return hash;
    }
}
