//***********************************************************************
//
//                              JMF.java
//
// Author(s) : Jean-luc LABLEE
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//
//**********************************************************************
package fis;

import java.awt.*;
import java.util.*;
import java.awt.geom.*;
import util.polygon2d.*;
import ptolemy.plot.*;

//------------------------------------------------------------------------------
/**
 * Classe de base des SEF
 */
public class JMF
{
	private long ptr; /**< pointeur vers classe native MF */
	//------------------------------------------------------------------------------
	public final static int TRIANGULAR = 0;  /**< SEF triangulaire. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int TRAPEZOIDAL = 1;  /**< SEF trapezoidal. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int SEMITRAPEZOIDALINF = 2;  /**< SEF trapezoidal inferieur. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int SEMITRAPEZOIDALSUP = 3;  /**< SEF trapezoidal superieur. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int DISCRETE = 4;  /**< SEF discret. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int UNIVERSAL = 5;  /**< SEF universel. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int GAUSSIAN = 6;  /**< SEF gaussien. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int DOOR = 7;  /**< SEF porte. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int SINUS = 8;  /**< SEF sinus. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int SINUSINF = 9;  /**< SEF sinus inferieur. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int SINUSSUP = 10;  /**< SEF sinus superieur. \warning il faut respecter l'ordre des types de SEF retournes par la fonction jnifis#MFType() */
	public final static int POSSIBILITY_DISTRIBUTION = 11;
	//------------------------------------------------------------------------------
	protected JMF( long ptr ) { this.ptr = ptr; }
	//------------------------------------------------------------------------------
	public JMF NewMF() throws Throwable
	{
		switch( this.GetType() )
		{
			case JMF.TRIANGULAR: return new JMFTriangular(this);
			case JMF.TRAPEZOIDAL:  return new JMFTrapezoidal(this);
			case JMF.SEMITRAPEZOIDALINF: return new JMFSemiTrapezoidalInf(this);
			case JMF.SEMITRAPEZOIDALSUP: return new JMFSemiTrapezoidalSup(this);
			case JMF.DISCRETE: return new JMFDiscrete(this);
			case JMF.UNIVERSAL: return new JMFUniversal(this);
			case JMF.GAUSSIAN: return new JMFGaussian(this);
			case JMF.DOOR: return new JMFDoor(this);
			case JMF.SINUS: return new JMFSinus(this);
			case JMF.SINUSINF: return new JMFSinusInf(this);
			case JMF.SINUSSUP: return new JMFSinusSup(this);
			case JMF.POSSIBILITY_DISTRIBUTION: return new JMFDPoss(this);
		}
		throw new Exception(LocaleFIS.GetString("UnknownMFType") + ": " + this.GetType());
	}
	//------------------------------------------------------------------------------
	public void Delete() throws Throwable { jnifis.DeleteMF(ptr); ptr = 0; }
	//------------------------------------------------------------------------------
	public long Ptr() { return ptr; }
	//------------------------------------------------------------------------------
	public String GetName() throws Throwable { return jnifis.GetMFName(ptr); }
	//------------------------------------------------------------------------------
	public void SetName( String name ) throws Throwable { jnifis.SetMFName(ptr, name); }
	//------------------------------------------------------------------------------
	public int GetType() throws Throwable
	{
		String type = jnifis.GetMFType(ptr);
		String [] MF_type = JMF.MFType();
		for( int i=0 ; i<MF_type.length ; i++ )
			if( type.equals(MF_type[i]) )
				return i;
		throw new Exception(LocaleFIS.GetString("UnknownMFType") + ": " + type);
	}
	//------------------------------------------------------------------------------
	public double [] GetParams() throws Throwable { return jnifis.GetMFParam(ptr); }
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set )  throws Throwable {}
	//------------------------------------------------------------------------------
	public static String [] MFType() throws Throwable { return jnifis.MFType(); }
	//------------------------------------------------------------------------------
	public double [] Support() throws Throwable { return jnifis.MFSupport(ptr); }
	//------------------------------------------------------------------------------
	public double [] NoyauAlpha( double alpha ) throws Throwable { return jnifis.MFAlphaKernel(ptr, alpha); }
	//------------------------------------------------------------------------------
	public void DrawMFDPoss( FISPlot plot, Graphics2D g2d, Vector<Double> dPoss) throws Throwable {}
	//------------------------------------------------------------------------------
	public void DrawAlphaKernel( FISPlot pb, Graphics2D g2d, double alpha )  throws Throwable
	{
		Point2D [] result = new Point2D[4];
		double [] support = this.Support();
		double [] noyau_alpha = this.NoyauAlpha(alpha);

		g2d.setColor(Color.red);
		if( pb.getPlotRectangle().height > 0 )
		{
			Polygon2D.Double poly = new Polygon2D.Double(pb.XtoPixel(noyau_alpha[0]),pb.YtoPixel(alpha));
			poly.lineTo(pb.XtoPixel(noyau_alpha[1]),pb.YtoPixel(alpha));
			poly.lineTo(pb.XtoPixel(support[1]),pb.YtoPixel(0));
			poly.lineTo(pb.XtoPixel(support[0]),pb.YtoPixel(0));
			g2d.fill( poly );
		}
	}
}



//------------------------------------------------------------------------------
class JMFTriangular extends JMF
{
	//------------------------------------------------------------------------------
	public JMFTriangular( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFTriangular( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFTriangular(name, params[0], params[1], params[2]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, params[0], 0, false);
		plot.addPoint(data_set, params[1], 1, true);
		plot.addPoint(data_set, params[2], 0, true);
	}
}



//------------------------------------------------------------------------------
class JMFTrapezoidal extends JMF
{
	//------------------------------------------------------------------------------
	public JMFTrapezoidal( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFTrapezoidal( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFTrapezoidal(name, params[0], params[1], params[2], params[3]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, params[0], 0, false);
		plot.addPoint(data_set, params[1], 1, true);
		plot.addPoint(data_set, params[2], 1, true);
		plot.addPoint(data_set, params[3], 0, true);
	}
}



//------------------------------------------------------------------------------
class JMFSemiTrapezoidalInf extends JMF
{
	//------------------------------------------------------------------------------
	public JMFSemiTrapezoidalInf( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFSemiTrapezoidalInf( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFSemiTrapezoidalInf(name, params[0], params[1], params[2]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, plot.GetXMin(), 1, false);
		plot.addPoint(data_set, params[1], 1, true);
		plot.addPoint(data_set, params[2], 0, true);
	}
}



//------------------------------------------------------------------------------
class JMFSemiTrapezoidalSup extends JMF
{
	//------------------------------------------------------------------------------
	public JMFSemiTrapezoidalSup( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFSemiTrapezoidalSup( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFSemiTrapezoidalSup(name, params[0], params[1], params[2]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, params[0], 0, false);
		plot.addPoint(data_set, params[1], 1, true);
		plot.addPoint(data_set, plot.GetXMax(), 1, true);
	}
}



//------------------------------------------------------------------------------
class JMFDiscrete extends JMF
{
	//------------------------------------------------------------------------------
	public JMFDiscrete( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFDiscrete( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFDiscrete(name, params) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		for( int i=0 ; i<params.length ; i++ )
		{
			plot.addPoint(data_set, params[i], 0, false);
			plot.addPoint(data_set, params[i], 1, true);
		}
	}
	//------------------------------------------------------------------------------
	public void DrawAlphaKernel( FISPlot pb, Graphics2D g2d, double alpha )  throws Throwable
	{
		g2d.setColor(Color.blue);
		double [] params = this.GetParams();
		for( int i=0 ; i<params.length ; i++ )
			g2d.draw( new Line2D.Double(pb.XtoPixel(params[i]), pb.YtoPixel(0), pb.XtoPixel(params[i]), pb.YtoPixel(alpha)) );


	}
}



//------------------------------------------------------------------------------
class JMFUniversal extends JMF
{
	//------------------------------------------------------------------------------
	public JMFUniversal( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFUniversal( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFUniversal(name, params[0], params[1]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, params[0], 1, false);
		plot.addPoint(data_set, params[1], 1, true);
	}
}



//------------------------------------------------------------------------------
class JMFGaussian extends JMF
{
	//------------------------------------------------------------------------------
	public JMFGaussian( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFGaussian( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFGaussian(name, params[1], params[0]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		int nb_points = 25;
		double [] x = new double [nb_points];
		double [] y = new double [nb_points];
		double dx = (5*params[1]) / (nb_points-1);
		plot.addPoint(data_set, params[0], 1, false);
		for( int i=0 ; i<nb_points ; i++ )
		{
			x[i] = i * dx;
			y[i] = Math.exp(-Math.pow(x[i], 2)/(2*Math.pow(params[1], 2)));
		}
		plot.addPoint(data_set, plot.GetXMax(), 0, false);
		for( int i=nb_points-1 ; i>=0 ; i-- )
			plot.addPoint(data_set, params[0]+x[i], y[i], true);
		for( int i=0 ; i<nb_points ; i++ )
			plot.addPoint(data_set, params[0]-x[i], y[i], true);
		plot.addPoint(data_set, plot.GetXMin(), 0, true);
	}
}

//------------------------------------------------------------------------------
class JMFDoor extends JMF
{
	//------------------------------------------------------------------------------
	public JMFDoor( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFDoor( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFDoor(name, params[0], params[1]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		plot.addPoint(data_set, params[0], 0, false);
		plot.addPoint(data_set, params[0], 1, true);
		plot.addPoint(data_set, params[1], 1, true);
		plot.addPoint(data_set, params[1], 0, true);
	}
}

//------------------------------------------------------------------------------
class JMFSinus extends JMF
{
	//------------------------------------------------------------------------------
	public JMFSinus( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFSinus( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFSinus(name, params[0], params[1]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		int nb_points = 25;
		double [] x = new double [nb_points];
		double [] y = new double [nb_points];
		double differ=params[1]-params[0];
		double EPSILON=1e-6;

		if (Math.abs(differ)<EPSILON) throw new Exception("ErrorInSinusParameters~range~too~small");
		double dx = (differ) / (nb_points-1);

		x[0]=params[0];
		y[0]=0;

		for( int i=1 ; i<nb_points ; i++ )
		{
			x[i] = x[i-1]+ dx;
			if (x[i]<params[0]) y[i] =0.0;
			if (x[i]>params[1]) y[i] =0.0;
			y[i] = Math.sin(Math.PI*(x[i]-params[0])/differ);
		}
		for( int i=0; i<nb_points; i++ )
			plot.addPoint(data_set, x[i], y[i], true);
	}
}

//------------------------------------------------------------------------------
class JMFSinusInf extends JMF {

	//------------------------------------------------------------------------------
	public JMFSinusInf( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFSinusInf( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFSinusInf(name, params[0], params[1]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		int nb_points = 25;
		double [] x = new double [nb_points];
		double [] y = new double [nb_points];
		double differ=params[1]-params[0];
		double EPSILON=1e-6;

		if (Math.abs(differ)<EPSILON) throw new Exception("ErrorInSinusParameters~range~too~small");
		double dx = (differ) / (nb_points-1);

		x[0]=params[0];
		y[0]=1;

		for( int i=1 ; i<nb_points ; i++ )
		{
			x[i] = x[i-1]+ dx;
			if (x[i]<params[0]) y[i] =1.0;
			if (x[i]>params[1]) y[i] =0.0;
			y[i] = Math.cos(Math.PI*0.5*(x[i]-params[0])/differ);
		}

		plot.addPoint(data_set, plot.GetXMin(), 1, false);
		for( int i=0; i<nb_points; i++ )
			plot.addPoint(data_set, x[i], y[i], true);
	}
}

//------------------------------------------------------------------------------
class JMFSinusSup extends JMF {

	//------------------------------------------------------------------------------
	public JMFSinusSup( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public JMFSinusSup( String name, double[] params ) throws Throwable
	{
		super( jnifis.NewMFSinusSup(name, params[0], params[1]) );
	}
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
		double [] params = GetParams();
		int nb_points = 25;
		double [] x = new double [nb_points];
		double [] y = new double [nb_points];
		double differ=params[1]-params[0];
		double EPSILON=1e-6;

		if (Math.abs(differ)<EPSILON) throw new Exception("ErrorInSinusParameters~range~too~small");
		double dx = (differ) / (nb_points-1);

		x[0]=params[0];
		y[0]=0;

		for( int i=1 ; i<nb_points ; i++ )
		{
			x[i] = x[i-1]+ dx;
			if (x[i]<params[0]) y[i] =0.0;
			if (x[i]>params[1]) y[i] =1.0;
			y[i] = Math.sin(Math.PI*0.5*(x[i]-params[0])/differ);
		}

		for( int i=0; i<nb_points; i++ )
			plot.addPoint(data_set, x[i], y[i], true);
		plot.addPoint(data_set, plot.GetXMax(), 1, true);
	}
}

//------------------------------------------------------------------------------
class JMFDPoss extends JMF
{
	//------------------------------------------------------------------------------
	public JMFDPoss( JMF sef ) { super( sef.Ptr() ); }
	//------------------------------------------------------------------------------
	public void Draw( FISPlot plot, int data_set ) throws Throwable
	{
	}
	//------------------------------------------------------------------------------
	public void DrawAlphaKernel( FISPlot pb, Graphics2D g2d, double alpha )  throws Throwable
	{
	}
	//------------------------------------------------------------------------------
	public void DrawMFDPoss( FISPlot plot, Graphics2D g2d, Vector<Double> dPoss ) throws Throwable
	{

		double [] params = GetParams();
		int j = 0;
		
		g2d.setColor(Color.blue);
		//transparence 80%
		g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.8f ));
	
		if( plot.getPlotRectangle().height > 0 && params.length > 0)
		{
			//creation du polygone avec 1 point
			Polygon2D.Double poly = new Polygon2D.Double(plot.XtoPixel(dPoss.elementAt(j).doubleValue()),plot.YtoPixel(dPoss.elementAt(j+1).doubleValue()));
			j++;
			for( int i = j+1 ; i<dPoss.size() ; i = i+2 )
			{
				poly.lineTo(plot.XtoPixel(dPoss.elementAt(i).doubleValue()),plot.YtoPixel(dPoss.elementAt(i+1).doubleValue()));
			}
			g2d.fill( poly );
		}
	}
	
	
}
