package fis.data.distance;

import static com.google.common.collect.ImmutableList.copyOf;
import static com.google.common.collect.Iterables.transform;
import static com.humaorie.dollar.Dollar.$;
import static org.apache.commons.io.FilenameUtils.concat;
import static org.apache.commons.io.FilenameUtils.getFullPath;

import java.beans.PropertyChangeEvent;
import java.io.File;

import util.algorithm.distance.minkowski.Minkowski;
import util.algorithm.distance.minkowski.MinkowskiChangeSupportDecorator;
import util.algorithm.distance.minkowski.MinkowskiModel;

import com.google.common.base.Function;
import com.google.common.collect.ImmutableList;
import com.jgoodies.binding.beans.Model;

import fis.JExtendedDataFile;
import fis.JFIS;
import fis.data.distance.input.Input;
import fis.data.distance.input.InputModel;

@SuppressWarnings("serial")
public class DistanceModel extends Model implements Distance {

	private final JFIS fis;
	private final JExtendedDataFile data;
	private final Minkowski minkowski = new MinkowskiModel();
	private MinkowskiChangeSupportDecorator minkowskiChangeSupportDecorator;
	private File outputFile;
	private final ImmutableList<Input> inputs;
	private boolean normalize = DEFAULT_NORMALIZE;

	public DistanceModel(JFIS fis, JExtendedDataFile data) throws Throwable {
		this.fis = fis;
		this.data = data;
		inputs = copyOf(getInputs(this.fis));
		outputFile = getOutputFile(this.fis, this.data);
	}

	private static File getOutputFile(JFIS fis, JExtendedDataFile data) throws Throwable {
		return new File(concat(getDataFullPath(data), String.format("%s.%s", fis.GetName(), DEFAULT_OUTPUT_FILE)));
	}

	private static String getDataFullPath(JExtendedDataFile data) {
		return getFullPath(data.GetFileName());
	}

	private static Iterable<Input> getInputs(JFIS fis) throws Throwable {
		return transform($(fis.NbInputs()), toInput(fis));
	}

	private static Function<Integer, Input> toInput(final JFIS fis) {
		return new Function<Integer, Input>() {

			@Override
			public Input apply(Integer inputIndex) {
				try {
					return new InputModel(fis.GetInput(inputIndex));
				} catch (Throwable e) {
					throw new RuntimeException(e);
				}
			}
		};
	}

	@Override
	public Minkowski getMinkowski() {
		return getMinkowskiChangeSupportDecorator();
	}

	private MinkowskiChangeSupportDecorator getMinkowskiChangeSupportDecorator() {
		return minkowskiChangeSupportDecorator != null ? minkowskiChangeSupportDecorator : (minkowskiChangeSupportDecorator = new MinkowskiChangeSupportDecorator(minkowski) {

			@Override
			public void beanChange(PropertyChangeEvent evt) {
				firePropertyChange(PROPERTY_MINKOWSKI_CHANGED, null, evt);
			}
		});
	}

	@Override
	public ImmutableList<Input> getInputs() {
		return inputs;
	}

	@Override
	public File getOutputFile() {
		return outputFile;
	}

	@Override
	public void setOutputFile(File outputFile) {
		firePropertyChange(PROPERTY_OUTPUT_FILE, this.outputFile, this.outputFile = outputFile);
	}

	public void release() {
		getMinkowskiChangeSupportDecorator().release();
	}

	@Override
	public boolean isNormalize() {
		return normalize;
	}

	@Override
	public void setNormalize(boolean normalize) {
		firePropertyChange(PROPERTY_NORMALIZE, this.normalize, this.normalize = normalize);
	}
}
