//***********************************************************************
//
//                              JDoubleMThumbSlider.java
//
// Author(s) : Jean-luc LABLEE
// FISPRO Version : 3.0 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Contact : fispro@ensam.inra.fr
// Last modification date:  December 25, 2005
//
//**********************************************************************
package util;

import javax.swing.DefaultBoundedRangeModel;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import util.mthumbslider.BasicMThumbSliderUI;
import util.mthumbslider.MThumbSlider;
import util.mthumbslider.MetalMThumbSliderUI;
import util.mthumbslider.MotifMThumbSliderUI;

/**
 * Classe slider a plusieurs curseurs avec valeurs en double
 * @author jean-luc.lablee
 */

public class JDoubleMThumbSlider extends MThumbSlider {
    protected int resolution;   // resolution du curseur

    private double[] doubleValue = null;  // valeurs des curseurs   

    private double[] previousDoubleValue = null; // valeurs precedentes des curseurs 

    protected double doubleMin; // valeur minimum du slider

    protected double doubleMax; // valeur maximum du slider

    // redirection des evenements de type ChangeListener exterieurs  la classe vers un EventListenerList local
    // la classe utilisant deja en interne un listener ChangeListener sur les curseurs
    private EventListenerList localListenerList = new EventListenerList();

    // croisement des curseurs. interdit par defaut
    protected boolean thumbCrossing = false;

    // distance min entre 2 curseurs quand le croisement est interdit
    protected int thumbMinimalDistance = 1;

    /**
     * Construit un slider  plusieurs curseurs
     * @param thumbNum nombre de curseurs
     */
    public JDoubleMThumbSlider(int thumbNum) {
        super(thumbNum);
        this.init();
    }

    /**
     * Construit un slider  2 curseurs
     */
    public JDoubleMThumbSlider() {
        super();
        this.init();
    }

    /**
     * Fonction d'init de la classe
     */
    private void init() {
        this.doubleValue = new double[this.thumbNum];
        this.previousDoubleValue = new double[this.thumbNum];

        // remplacement des modles de MThumbSlider
        // tous les modeles ont les memes resolution, min et max
        this.resolution = 100;
        for (int i = 0; i < this.thumbNum; i++) {
            final int fi = i;
            this.sliderModels[i] = new DefaultBoundedRangeModel(this.thumbNum != 1 ? i * (this.resolution / (this.thumbNum - 1)) : this.resolution / 2, 0, 0, this.resolution);
            this.sliderModels[i].addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent e) {
                    JDoubleMThumbSlider.this.sliderStateChanged(fi);
                }
            });
        }

        // valeurs par defaut
        this.doubleMin = this.sliderModels[0].getMinimum();
        this.doubleMax = this.sliderModels[0].getMaximum();
        for (int i = 0; i < this.thumbNum; i++)
            this.previousDoubleValue[i] = this.doubleValue[i] = this.sliderModels[i].getValue();
    }

    /**
     * Modifie les valeurs minimum et maximum du slider
     * @param min valeur minimum
     * @param max valeur maximum
     */
    public void setDoubleMinMax(double min, double max) {
        // les curseurs ont les memes min et max
        this.doubleMin = min;
        this.doubleMax = max;
        for (int i = 0; i < this.thumbNum; i++)
            this.previousDoubleValue[i] = this.doubleValue[i] = this.modelValueToDoubleValue(this.sliderModels[i].getValue());
    }
    
    /**
     * @return Renvoie doubleMax.
     */
    public double getDoubleMax() {
        return doubleMax;
    }

    /**
     * @return Renvoie doubleMin.
     */
    public double getDoubleMin() {
        return doubleMin;
    }

    /**
     * Converti une valeur du modle en valeur en double
     * @param modelValue valeur du modle
     * @return valeur en double
     */
    protected double modelValueToDoubleValue(int modelValue) {
        return JDoubleRound.round(this.doubleMin + (((double) modelValue / (double) this.resolution) * (this.doubleMax - this.doubleMin)));
    }

    /**
     * Converti une valeur en double en valeur du modle
     * @param doubleValue valeur en double
     * @return valeur du modle
     */
    protected int doubleValueToModelValue(double doubleValue) {
        return (int) JDoubleRound.round(((doubleValue - this.doubleMin) / (this.doubleMax - this.doubleMin)) * this.resolution);
        }

    /**
     * Fixe la valeur d'un curseur 
     * @param index index du curseur
     * @param doubleValue valeur
     */
    public void setDoubleValue(int index, double doubleValue) {
        this.previousDoubleValue[index] = this.doubleValue[index] = doubleValue;
        int modelValue = this.doubleValueToModelValue(doubleValue);
        this.setValueAt(modelValue, index);     // utilisation de setValueAt autorise ici
        // modelValue modifi par setValueAt (si croisement de curseurs interdit)
        if( this.getValueAt(index) != modelValue )
            this.previousDoubleValue[index] = this.doubleValue[index] = this.modelValueToDoubleValue(this.getValueAt(index));
    }

    /**
     * Renvoie la valeur d'un curseur
     * @param index index du curseur
     * @return valeur du curseur
     */
    public double getDoubleValue(int index) {
        return this.doubleValue[index];
    }

    /** 
     * @see util.mthumbslider.MThumbSlider#setValueAt(int, int)
     * @deprecated utiliser setDoubleValue()  la place
     */
    public void setValueAt(int n, int index) {
        if (this.thumbCrossing == true)
            super.setValueAt(n, index);
        else {
            // croisement des curseurs interdit
            int previousThumbValue = index == 0 ? 0 : this.getValueAt(index - 1) + this.thumbMinimalDistance;
            int nextThumbValue = index == (this.thumbNum - 1) ? this.resolution : this.getValueAt(index + 1) - this.thumbMinimalDistance;
            super.setValueAt(Math.min(Math.max(previousThumbValue, n), nextThumbValue), index);
        }
    }

    /* (non-Javadoc)
     * @see javax.swing.JComponent#updateUI()
     */
    public void updateUI() {
        updateLabelUIs();
        String lookAndFeelName = UIManager.getLookAndFeel().getName();
        if (lookAndFeelName.startsWith("Metal"))
            setUI(new MetalMThumbSliderUI(this));
        else if (lookAndFeelName.startsWith("CDE/"))
            setUI(new MotifMThumbSliderUI(this));
        else
            setUI(new BasicMThumbSliderUI(this));
    }

    /**
     * @param thumbCrossing
     *            thumbCrossing  dfinir.
     */
    public void setThumbCrossing(boolean thumbCrossing) {
        this.thumbCrossing = thumbCrossing;
    }

    /**
     * Appell lors du dplacement des curseurs
     * @param thumbIndex index du curseur
     */
    protected void sliderStateChanged(int thumbIndex) {
        // slider en ajustement
        if (this.getValueIsAdjusting()) {
            this.doubleValue[thumbIndex] = modelValueToDoubleValue(sliderModels[thumbIndex].getValue()); 
            // test si la valeur a chang
            if (this.doubleValue[thumbIndex] != this.previousDoubleValue[thumbIndex]) {
                this.previousDoubleValue[thumbIndex] = this.doubleValue[thumbIndex];
                this.fireLocalChangeEvent();
            }
        }

    }

    /* (non-Javadoc)
     * @see javax.swing.JSlider#addChangeListener(javax.swing.event.ChangeListener)
     */
    public void addChangeListener(ChangeListener cl) {
        // redirection des listeners de type ChangeListener vers le localListenerList
        this.localListenerList.add(ChangeListener.class, cl);
    }

    /* (non-Javadoc)
     * @see javax.swing.JSlider#getChangeListeners()
     */
    public ChangeListener[] getChangeListeners() {
        return (ChangeListener[]) this.localListenerList.getListeners(ChangeListener.class);
    }

    /* (non-Javadoc)
     * @see javax.swing.JSlider#removeChangeListener(javax.swing.event.ChangeListener)
     */
    public void removeChangeListener(ChangeListener cl) {
        this.localListenerList.remove(ChangeListener.class, cl);
    }

    /**
     * Appel des listeners locaux
     */
    protected void fireLocalChangeEvent() {
        ChangeListener[] listeners = this.getChangeListeners();
        for (int i = 0; i < listeners.length; i++)
            listeners[i].stateChanged(new ChangeEvent(this));
    }

}
