//***********************************************************************
//
//                              FuzzyTree.java
//
// Author(s) : Jean-luc LABLEE
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
//
//**********************************************************************
package fis;

import javax.swing.JTree;
import javax.swing.*;
import java.awt.*;
import javax.swing.tree.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.border.*;
import javax.swing.event.*;

//------------------------------------------------------------------------------
public class FuzzyTree extends JTree
  {
  private int NbClass;                      // Nombre de classes de l'arbre
  private FuzzyNode Root;                   // noeud racine de l'arbre
  private DefaultTreeModel TreeModel;       // Model de larbre
  private Vector Nodes = new Vector();      // vecteur des noeuds de l'arbre
  Vector VariablesGroups = new Vector();    // vecteur des groupes de variables
  boolean DisplayAllNodes = true;           // Affiche tous les noeuds
  private boolean Classif = false;          // arbre classif
  private boolean Regression = false;       // arbre regression
  private boolean FuzzyOutput = false;      // arbre flou
//------------------------------------------------------------------------------
  public FuzzyTree( String file_name ) throws FileNotFoundException, IOException, Exception
    {
    // ouverture du fichier file_name
    File f = new File(file_name);
    // creation d'un buffer de lecture du fichier f
    BufferedReader br = new BufferedReader( new InputStreamReader( new FileInputStream(f) ) );
    String buffer = br.readLine();
    Vector params = new Vector(); // parametres sur premiere ligne du fichier

    try
      {
      StringTokenizer st = new StringTokenizer( buffer, "," ); // decompose le buffer (separateur ,)
      this.NbClass = Integer.parseInt(st.nextToken().trim()); // lecture du nombre de classes
      int class_number = 1;
      while( st.countTokens() != 0 )
        {
        String class_name = new String(st.nextToken().trim());
        params.add(class_name.equals("") == true ? new String("Class "+class_number) : class_name);  // noms des classes (affecte un nom par defaut si celui-ci n'est pas present dans le fichier arbre)
        class_number++;
        }
      }
    catch( NumberFormatException e ) { throw new Exception( LocaleFIS.GetString("ErrorInTreeFile") + ": " + file_name + "\n" + LocaleFIS.GetString("UnexpectedValueForNumberOfClass") + " : " + buffer); }

    // determination du type d'arbre
    if( this.NbClass < 0 )    // arbre flou
      {
      this.FuzzyOutput = true;
      this.NbClass = Math.abs(this.NbClass);
      }
    else if( this.NbClass == 0 )  // arbre regression
      this.Regression = true;
    else if( this.NbClass > 0 )   // arbre classification
      this.Classif = true;

    if( (buffer = br.readLine()) != null )
      {
      try
        {
        // init de la racine de l'arbre
        this.Root = this.CreateNode(buffer);  // creation du noeud racine
        this.Root.InitVariablesGroups(this.VariablesGroups, params);  // Init des noms des variables affichables
        this.TreeModel = new DefaultTreeModel(this.Root);  // creation du model de l'arbre
        while( (buffer = br.readLine()) != null ) // tant que le fichier n'est pas entierement lu
          {
          FuzzyNode new_node = null;
          new_node = this.CreateNode(buffer); // creation d'un nouveau noeud
          DefaultMutableTreeNode parent_node = this.ParentNode(new_node); // recherche du noeud parent du nouveau noeud
          this.TreeModel.insertNodeInto(new_node, parent_node, parent_node.getChildCount()); // insertion du nouveau noeud dans le noeud parent
          }
        }
      catch( Exception e ) { throw new Exception( LocaleFIS.GetString("ErrorInTreeFile") + ": " + file_name + "\n" + e.getMessage() ); }
      }
    br.close(); // fermeture du fichier arbre
    this.setModel(this.TreeModel); // affectation du model a l'arbre
    }
//------------------------------------------------------------------------------
// recherche et retourne le noeud parent du noeud new_node
  public FuzzyNode ParentNode( FuzzyNode new_node ) throws Exception
    {
    for( int i=0 ; i<this.Nodes.size() ; i++ )    // parmis tous les noeud enregistres
      if( ((FuzzyNode)this.Nodes.elementAt(i)).Node() == new_node.Parent() )
        return (FuzzyNode)this.Nodes.elementAt(i);
    throw new Exception( LocaleFIS.GetString("ParentNode") + ": " + Integer.toString(new_node.Parent()) + " " + LocaleFIS.GetString("NotFoundForNode") + ": " + Integer.toString(new_node.Node()) );
    }
//------------------------------------------------------------------------------
// cre et retourne un noeud en fonction du type d'arbre
  private FuzzyNode CreateNode( String buffer ) throws Exception
    {
    FuzzyNode new_node = null;
    if( this.Classif == true )
      new_node = new FuzzyNodeClassif( buffer, this.NbClass );  // creation d'un noeud Classif
    if( this.Regression == true )
      new_node = new FuzzyNodeRegression( buffer ); // creation d'un noeud Regression
    if( this.FuzzyOutput == true )
      new_node = new FuzzyNodeFuzzyOutput( buffer, this.NbClass );  // creation d'un noeud flou
    this.Nodes.add(new_node);   // ajout du noeud dans le vecteur des noeuds
    return new_node;
    }
//------------------------------------------------------------------------------
  public int NbClass() { return this.NbClass; }
//------------------------------------------------------------------------------
  public boolean Classif() { return this.Classif; }
//------------------------------------------------------------------------------
  public boolean Regression() { return this.Regression; }
//------------------------------------------------------------------------------
  public boolean FuzzyOutput() { return this.FuzzyOutput; }
  }





//------------------------------------------------------------------------------
// classe de base abstraite d'un noeud
abstract class FuzzyNode extends DefaultMutableTreeNode
  {
  private int Node;                 // numero du noeud
  private int Parent;               // numero du noeud parent
  protected Vector Variables = new Vector();  // vecteur des variables du noeud
  protected StringTokenizer st;
//------------------------------------------------------------------------------
  public FuzzyNode( String buffer ) throws Exception
    {
    st = new StringTokenizer(buffer, "," ); // decompose le buffer
    if( st.countTokens() < 7 )
      throw new Exception( LocaleFIS.GetString("UnexpectedFormat") + " : " + buffer );

    // lecture du noeud (partie commune a tous les types de noeuds)
    this.Node = Integer.parseInt(st.nextToken().trim());  // lecture du numero du noeud
    this.Variables.add(new String( st.nextToken().trim() + " " + st.nextToken().trim() ));  // lecture du label variable + label sef
    this.Parent = Integer.parseInt(st.nextToken().trim());  // lecture du numero du noeud parent
    st.nextToken(); // saute le label feuille

    this.Variables.add( new Integer(st.nextToken().trim()) ); // lecture du nombre d'exemples
    this.Variables.add( new Double(st.nextToken().trim()) );  // lecture entropie

    this.setUserObject(this.Variables.elementAt(0));
    }
//------------------------------------------------------------------------------
// retourne le numero du noeud parent
  public int Parent() { return this.Parent; }
//------------------------------------------------------------------------------
// retourne le numero du noeud
  public int Node() { return this.Node; }
//------------------------------------------------------------------------------
// rempli le vecteur display_variables avec le nom des variables
  public void InitVariablesGroups( Vector display_variables, Vector params )
    {
    display_variables.add(new DisplayVariableGroup(new String("Samples")));
    }
  }







//------------------------------------------------------------------------------
// classe noeud classif
class FuzzyNodeClassif extends FuzzyNode
  {
  private int nb_class; // nombre de classes pour la classif
//------------------------------------------------------------------------------
  public FuzzyNodeClassif(  String buffer, int nb_class ) throws Exception
    {
    super( buffer );
    this.nb_class = nb_class;

    // lecture du noeud
    this.Variables.add(new Integer(st.nextToken().trim()) );  // lecture de la classe majoritaire
    for( int i=0 ; i<nb_class ; i++ )
      this.Variables.add(new Double(st.nextToken().trim()));  // lecture des effectifs par classe
    }
//------------------------------------------------------------------------------
// rempli le vecteur display_variables avec le nom des variables
  public void InitVariablesGroups( Vector display_variables, Vector params )
    {
    super.InitVariablesGroups(display_variables, params);
    display_variables.add(new DisplayVariableGroup(new String("Entropie")));
    // le nom de la variable du groupe MajorityClass est affiche sur 2 lignes
    display_variables.add(new DisplayVariableGroup(new String("MajorityClass"), new String("Majority\nClass")));
    // les effectifs par classe sont geres par un groupe de variables ClassEffectif
    DisplayVariableGroup new_dv = new DisplayVariableGroup(new String("ClassEffectif"), Integer.toString(1));
    for( int i=1 ; i<this.nb_class ; i++ )
      new_dv.AddVariableName(Integer.toString(i+1));
    display_variables.add(new_dv);
    }
  }



//------------------------------------------------------------------------------
// classe noeud regression
class FuzzyNodeRegression extends FuzzyNode
  {
//------------------------------------------------------------------------------
  public FuzzyNodeRegression( String buffer ) throws Exception
    {
    super( buffer );

    // lecture du noeud
    this.Variables.add(new Double(st.nextToken().trim()) ); // lecture moyenne
    this.Variables.add(new Double(st.nextToken().trim()) ); // lecture ecart-type
    }
//------------------------------------------------------------------------------
// rempli le vecteur display_variables avec le nom des variables
  public void InitVariablesGroups( Vector display_variables, Vector params )
    {
    super.InitVariablesGroups(display_variables, params);
    display_variables.add(new DisplayVariableGroup(new String("Deviance")));
    display_variables.add(new DisplayVariableGroup(new String("Mean")));
    display_variables.add(new DisplayVariableGroup(new String("StandardDeviation")));
    }
  }


//------------------------------------------------------------------------------
// classe noeud flou
class FuzzyNodeFuzzyOutput extends FuzzyNode
  {
  private int nb_class; // nombre de classe (en fait nombre de sef dans ce cas)
//------------------------------------------------------------------------------
  public FuzzyNodeFuzzyOutput(  String buffer, int nb_class ) throws Exception
    {
    super( buffer );
    this.nb_class = nb_class;

    // lecture du noeud
    this.Variables.add(new String(st.nextToken().trim()));  // lecture sef majoritaire
    for( int i=0 ; i<nb_class ; i++ )
      this.Variables.add(new Double(st.nextToken().trim()));  // lecture de l'effectif par sef
    }
//------------------------------------------------------------------------------
// rempli le vecteur display_variables avec le nom des variables
  public void InitVariablesGroups( Vector display_variables, Vector params )
    {
    super.InitVariablesGroups(display_variables, params);
    display_variables.add(new DisplayVariableGroup(new String("Entropie")));
    // le nom de la variable du groupe MajorityClass est affiche sur 2 lignes
    display_variables.add(new DisplayVariableGroup(new String("MajorityClass"), new String("Majority\nClass")));
    // les effectifs par sef sont geres par un groupe de variables ClassEffectif
    DisplayVariableGroup new_dv = new DisplayVariableGroup(new String("ClassEffectif"), (String)params.elementAt(0));
    for( int i=1 ; i<this.nb_class ; i++ )
      new_dv.AddVariableName((String)params.elementAt(i));
    display_variables.add(new_dv);
    }
  }

//------------------------------------------------------------------------------
// classe de gestion des noms et de l'affichage des groupes de variables
class DisplayVariableGroup
  {
  private String GroupName; // nom du groupe de variable
  private boolean Display;  // flag groupe de variable affichable
  private Vector VariablesNames = new Vector();  // vecteur des noms de variables
//------------------------------------------------------------------------------
// a utiliser lorsque le nom de la variable est le meme que le nom du groupe
  public DisplayVariableGroup( String group_name )
    {
    this.GroupName = group_name;
    this.Display = true;
    this.VariablesNames.add(group_name);
    }
//------------------------------------------------------------------------------
// a utiliser lorsque le nom de la variable est different du nom du groupe
  public DisplayVariableGroup( String group_name, String name )
    {
    this.GroupName = group_name;
    this.Display = true;
    this.VariablesNames.add(name);
    }
//------------------------------------------------------------------------------
// retourne le nom du groupe de variable
  String GroupName() { return this.GroupName; }
//------------------------------------------------------------------------------
// retourne le flag affichable
  boolean Display() { return this.Display; }
//------------------------------------------------------------------------------
// modifie le flag affichable
  void Display( boolean display ) { this.Display = display; }
//------------------------------------------------------------------------------
// ajoute une variable au groupe
  void AddVariableName( String name ) { this.VariablesNames.add(name); }
//------------------------------------------------------------------------------
// retourne le nom de la variable index
  String VariableName( int index ) { return (String)this.VariablesNames.elementAt(index); }
//------------------------------------------------------------------------------
// retourne la taille du groupe
  int VariablesNamesSize() { return this.VariablesNames.size(); }
  }


